@@ -38,7 +38,7 @@ ERROR_T libpng_version(char *buffer) {
     /* retrieves the libpng version value and then extracts
     the multiple parts from it */
     version = png_access_version_number();
-    sprintf(version_s, "%06d", version);
+    sprintf(version_s, "%06d", (int) version);
 
     memcpy(major_s, &version_s[0], 2);
     memcpy(medium_s, &version_s[2], 2);
@@ -81,8 +81,8 @@ ERROR_T read_png(char *file_name, char demultiply, struct pcv_image *image) {
     png_uint_32 buffer_size;
     png_uint_32 row_size;
 
-    /* opens the file and tests for it being a PNG, this is required
-    to avoid possible problems while handling improper files */
+    /* opens the file and tests for proper opening, this is required
+    to avoid possible problems while handling improper file reading */
 #ifdef _MSC_VER
     wchar_t file_name_w[1024];
     swprintf(file_name_w, 1024, L"\\\\?\\%hs", file_name);
@@ -93,6 +93,9 @@ ERROR_T read_png(char *file_name, char demultiply, struct pcv_image *image) {
     if(!fp) {
         RAISE_F("[read_png] File %s could not be opened for reading", file_name);
     }
+
+    /* tries to read the PNG signature from the file and in case it's
+    not valid raises an error indicating so */
     count = fread(header, 1, 8, fp);
     if(png_sig_cmp((void *) header, 0, 8) || count != 8) {
         RAISE_F("[read_png] File %s is not recognized as a PNG file", file_name);
@@ -665,7 +668,13 @@ ERROR_T duplicate_image(struct pcv_image *origin, struct pcv_image *target) {
     NORMAL;
 }
 
-ERROR_T compose_images(char *base_path, char *algorithm, params *params, char *background) {
+ERROR_T compose_images(
+    char *base_path,
+    char *algorithm,
+    params *params,
+    char *background,
+    struct benchmark *benchmark
+) {
     return compose_images_extra(
         base_path,
         algorithm,
@@ -673,7 +682,8 @@ ERROR_T compose_images(char *base_path, char *algorithm, params *params, char *b
         background,
         Z_BEST_SPEED,
         PNG_FILTER_NONE,
-        FALSE
+        FALSE,
+        benchmark
     );
 }
 
@@ -684,36 +694,80 @@ ERROR_T compose_images_extra(
     char *background,
     int compression,
     int filter,
-    char use_opencl
+    char use_opencl,
+    struct benchmark *benchmark
 ) {
     char path[1024];
     char name[1024];
+    float start_time;
     struct pcv_image bottom, top, final;
     char demultiply = is_multiplied(algorithm);
 
-    VALIDATE(read_png(join_path(base_path, "sole.png", path), demultiply, &bottom));
-    VALIDATE(read_png(join_path(base_path, "back.png", path), demultiply, &top));
-    VALIDATE(blend_images_extra(&bottom, &top, algorithm, params, use_opencl));
-    VALIDATE(release_image(&top));
+    BENCHMARK(
+        benchmark, start_time, benchmark->read_png_time,
+        VALIDATE(read_png(join_path(base_path, "sole.png", path), demultiply, &bottom));
+        VALIDATE(read_png(join_path(base_path, "back.png", path), demultiply, &top))
+    );
+    BENCHMARK(
+        benchmark, start_time, benchmark->blend_time,
+        VALIDATE(blend_images_extra(&bottom, &top, algorithm, params, use_opencl))
+    );
+    BENCHMARK(
+        benchmark, start_time, benchmark->read_png_time,
+        VALIDATE(release_image(&top))
+    );
 
-    VALIDATE(read_png(join_path(base_path, "front.png", path), demultiply, &top));
-    VALIDATE(blend_images_extra(&bottom, &top, algorithm, params, use_opencl));
-    VALIDATE(release_image(&top));
+    BENCHMARK(
+        benchmark, start_time, benchmark->read_png_time,
+        VALIDATE(read_png(join_path(base_path, "front.png", path), demultiply, &top))
+    );
+    BENCHMARK(
+        benchmark, start_time, benchmark->blend_time,
+        VALIDATE(blend_images_extra(&bottom, &top, algorithm, params, use_opencl))
+    );
+    BENCHMARK(
+        benchmark, start_time, benchmark->read_png_time,
+        VALIDATE(release_image(&top))
+    );
 
-    VALIDATE(read_png(join_path(base_path, "shoelace.png", path), demultiply, &top));
-    VALIDATE(blend_images_extra(&bottom, &top, algorithm, params, use_opencl));
-    VALIDATE(release_image(&top));
+    BENCHMARK(
+        benchmark, start_time, benchmark->read_png_time,
+        VALIDATE(read_png(join_path(base_path, "shoelace.png", path), demultiply, &top))
+    );
+    BENCHMARK(
+        benchmark, start_time, benchmark->blend_time,
+        VALIDATE(blend_images_extra(&bottom, &top, algorithm, params, use_opencl))
+    );
+    BENCHMARK(
+        benchmark, start_time, benchmark->read_png_time,
+        VALIDATE(release_image(&top))
+    );
 
-    if(demultiply) { multiply_image(&bottom); }
+    BENCHMARK(
+        benchmark, start_time, benchmark->blend_time,
+        if(demultiply) { multiply_image(&bottom); }
+    );
 
-    sprintf(name, "background_%s.png", background);
-    VALIDATE(read_png(join_path(base_path, name, path), FALSE, &final));
-    VALIDATE(blend_images_extra(&final, &bottom, algorithm, params, use_opencl));
-    VALIDATE(release_image(&bottom));
+    BENCHMARK(
+        benchmark, start_time, benchmark->read_png_time,
+        sprintf(name, "background_%s.png", background);
+        VALIDATE(read_png(join_path(base_path, name, path), FALSE, &final));
+    );
+    BENCHMARK(
+        benchmark, start_time, benchmark->blend_time,
+        VALIDATE(blend_images_extra(&final, &bottom, algorithm, params, use_opencl));
+    );
+    BENCHMARK(
+        benchmark, start_time, benchmark->read_png_time,
+        VALIDATE(release_image(&bottom))
+    );
 
-    sprintf(name, "result_%s_%s.png", algorithm, background);
-    VALIDATE(write_png_extra(&final, FALSE, join_path(base_path, name, path), compression, filter));
-    VALIDATE(release_image(&final));
+    BENCHMARK(
+        benchmark, start_time, benchmark->write_png_time,
+        sprintf(name, "result_%s_%s_%s.png", algorithm, background, use_opencl ? "opencl" : "cpu");
+        VALIDATE(write_png_extra(&final, FALSE, join_path(base_path, name, path), compression, filter));
+        VALIDATE(release_image(&final))
+    );
 
     NORMAL;
 }
@@ -721,42 +775,42 @@ ERROR_T compose_images_extra(
 ERROR_T pcompose(int argc, char **argv) {
     if(argc != 3) { RAISE_M("Usage: pconvert compose <directory>"); }
 
-    VALIDATE(compose_images(argv[2], "alpha", NULL, "alpha"));
-    VALIDATE(compose_images(argv[2], "alpha", NULL, "white"));
-    VALIDATE(compose_images(argv[2], "alpha", NULL, "blue"));
-    VALIDATE(compose_images(argv[2], "alpha", NULL, "texture"));
-    VALIDATE(compose_images(argv[2], "multiplicative", NULL, "alpha"));
-    VALIDATE(compose_images(argv[2], "multiplicative", NULL, "white"));
-    VALIDATE(compose_images(argv[2], "multiplicative", NULL, "blue"));
-    VALIDATE(compose_images(argv[2], "multiplicative", NULL, "texture"));
-    VALIDATE(compose_images(argv[2], "source_over", NULL, "alpha"));
-    VALIDATE(compose_images(argv[2], "source_over", NULL, "white"));
-    VALIDATE(compose_images(argv[2], "source_over", NULL, "blue"));
-    VALIDATE(compose_images(argv[2], "source_over", NULL, "texture"));
-    VALIDATE(compose_images(argv[2], "destination_over", NULL, "alpha"));
-    VALIDATE(compose_images(argv[2], "destination_over", NULL, "white"));
-    VALIDATE(compose_images(argv[2], "destination_over", NULL, "blue"));
-    VALIDATE(compose_images(argv[2], "destination_over", NULL, "texture"));
-    VALIDATE(compose_images(argv[2], "first_top", NULL, "alpha"));
-    VALIDATE(compose_images(argv[2], "first_top", NULL, "white"));
-    VALIDATE(compose_images(argv[2], "first_top", NULL, "blue"));
-    VALIDATE(compose_images(argv[2], "first_top", NULL, "texture"));
-    VALIDATE(compose_images(argv[2], "first_bottom", NULL, "alpha"));
-    VALIDATE(compose_images(argv[2], "first_bottom", NULL, "white"));
-    VALIDATE(compose_images(argv[2], "first_bottom", NULL, "blue"));
-    VALIDATE(compose_images(argv[2], "first_bottom", NULL, "texture"));
-    VALIDATE(compose_images(argv[2], "disjoint_over", NULL, "alpha"));
-    VALIDATE(compose_images(argv[2], "disjoint_over", NULL, "white"));
-    VALIDATE(compose_images(argv[2], "disjoint_over", NULL, "blue"));
-    VALIDATE(compose_images(argv[2], "disjoint_over", NULL, "texture"));
-    VALIDATE(compose_images(argv[2], "disjoint_under", NULL, "alpha"));
-    VALIDATE(compose_images(argv[2], "disjoint_under", NULL, "white"));
-    VALIDATE(compose_images(argv[2], "disjoint_under", NULL, "blue"));
-    VALIDATE(compose_images(argv[2], "disjoint_under", NULL, "texture"));
-    VALIDATE(compose_images(argv[2], "disjoint_debug", NULL, "alpha"));
-    VALIDATE(compose_images(argv[2], "disjoint_debug", NULL, "white"));
-    VALIDATE(compose_images(argv[2], "disjoint_debug", NULL, "blue"));
-    VALIDATE(compose_images(argv[2], "disjoint_debug", NULL, "texture"));
+    VALIDATE(compose_images(argv[2], "alpha", NULL, "alpha", NULL));
+    VALIDATE(compose_images(argv[2], "alpha", NULL, "white", NULL));
+    VALIDATE(compose_images(argv[2], "alpha", NULL, "blue", NULL));
+    VALIDATE(compose_images(argv[2], "alpha", NULL, "texture", NULL));
+    VALIDATE(compose_images(argv[2], "multiplicative", NULL, "alpha", NULL));
+    VALIDATE(compose_images(argv[2], "multiplicative", NULL, "white", NULL));
+    VALIDATE(compose_images(argv[2], "multiplicative", NULL, "blue", NULL));
+    VALIDATE(compose_images(argv[2], "multiplicative", NULL, "texture", NULL));
+    VALIDATE(compose_images(argv[2], "source_over", NULL, "alpha", NULL));
+    VALIDATE(compose_images(argv[2], "source_over", NULL, "white", NULL));
+    VALIDATE(compose_images(argv[2], "source_over", NULL, "blue", NULL));
+    VALIDATE(compose_images(argv[2], "source_over", NULL, "texture", NULL));
+    VALIDATE(compose_images(argv[2], "destination_over", NULL, "alpha", NULL));
+    VALIDATE(compose_images(argv[2], "destination_over", NULL, "white", NULL));
+    VALIDATE(compose_images(argv[2], "destination_over", NULL, "blue", NULL));
+    VALIDATE(compose_images(argv[2], "destination_over", NULL, "texture", NULL));
+    VALIDATE(compose_images(argv[2], "first_top", NULL, "alpha", NULL));
+    VALIDATE(compose_images(argv[2], "first_top", NULL, "white", NULL));
+    VALIDATE(compose_images(argv[2], "first_top", NULL, "blue", NULL));
+    VALIDATE(compose_images(argv[2], "first_top", NULL, "texture", NULL));
+    VALIDATE(compose_images(argv[2], "first_bottom", NULL, "alpha", NULL));
+    VALIDATE(compose_images(argv[2], "first_bottom", NULL, "white", NULL));
+    VALIDATE(compose_images(argv[2], "first_bottom", NULL, "blue", NULL));
+    VALIDATE(compose_images(argv[2], "first_bottom", NULL, "texture", NULL));
+    VALIDATE(compose_images(argv[2], "disjoint_over", NULL, "alpha", NULL));
+    VALIDATE(compose_images(argv[2], "disjoint_over", NULL, "white", NULL));
+    VALIDATE(compose_images(argv[2], "disjoint_over", NULL, "blue", NULL));
+    VALIDATE(compose_images(argv[2], "disjoint_over", NULL, "texture", NULL));
+    VALIDATE(compose_images(argv[2], "disjoint_under", NULL, "alpha", NULL));
+    VALIDATE(compose_images(argv[2], "disjoint_under", NULL, "white", NULL));
+    VALIDATE(compose_images(argv[2], "disjoint_under", NULL, "blue", NULL));
+    VALIDATE(compose_images(argv[2], "disjoint_under", NULL, "texture", NULL));
+    VALIDATE(compose_images(argv[2], "disjoint_debug", NULL, "alpha", NULL));
+    VALIDATE(compose_images(argv[2], "disjoint_debug", NULL, "white", NULL));
+    VALIDATE(compose_images(argv[2], "disjoint_debug", NULL, "blue", NULL));
+    VALIDATE(compose_images(argv[2], "disjoint_debug", NULL, "texture", NULL));
 
     NORMAL;
 }
@@ -781,13 +835,17 @@ float pbenchmark_algorithm(
     char *background,
     int compression,
     int filter,
-    char use_opencl
+    char use_opencl,
+    struct benchmark *benchmark
 ) {
     float start_time;
     float end_time;
     float time_elapsed;
     start_time = (float) clock() / CLOCKS_PER_SEC;
-    EXCEPT_P(compose_images_extra(base_path, algorithm, params, background, compression, filter, use_opencl));
+    EXCEPT_S(compose_images_extra(
+        base_path, algorithm, params, background, compression,
+        filter, use_opencl, benchmark
+    ));
     end_time = (float) clock() / CLOCKS_PER_SEC;
     time_elapsed = end_time - start_time;
     return time_elapsed;
@@ -796,11 +854,33 @@ float pbenchmark_algorithm(
 ERROR_T pbenchmark(int argc, char **argv) {
     #define ALGORITHMS_SIZE 5
     #define COMPRESSION_SIZE 3
-    size_t index, index_j, index_k;
-    float time;
-    char *algorithms[ALGORITHMS_SIZE] = { "multiplicative", "source_over", "alpha", "disjoint_over", "disjoint_under" };
-    int compression[COMPRESSION_SIZE] = { Z_NO_COMPRESSION, Z_BEST_SPEED, Z_BEST_COMPRESSION };
-    char *compression_s[COMPRESSION_SIZE] = { "Z_NO_COMPRESSION", "Z_BEST_SPEED", "Z_BEST_COMPRESSION" };
+
+    struct benchmark benchmark, benchmark_total = { 0, 0, 0};
+    size_t index, index_j, index_k, index_c;
+    float time, result, total = 0.0f;
+    char label[128];
+    char time_s[64];
+    char count = 1;
+    char is_success = TRUE;
+    char details = TRUE;
+
+    char *algorithms[ALGORITHMS_SIZE] = {
+        "multiplicative",
+        "source_over",
+        "alpha",
+        "disjoint_over",
+        "disjoint_under"
+    };
+    int compression[COMPRESSION_SIZE] = {
+        Z_NO_COMPRESSION,
+        Z_BEST_SPEED,
+        Z_BEST_COMPRESSION
+    };
+    char *compression_s[COMPRESSION_SIZE] = {
+        "Z_NO_COMPRESSION",
+        "Z_BEST_SPEED",
+        "Z_BEST_COMPRESSION"
+    };
 
 #ifdef PCONVERT_OPENCL
     #define OPENCL_SIZE 2
@@ -812,18 +892,61 @@ ERROR_T pbenchmark(int argc, char **argv) {
     char *use_opencl_s[OPENCL_SIZE] = { "CPU" };
 #endif
 
-    if(argc != 3) { RAISE_M("Usage: pconvert benchmark <directory>"); }
+    if(argc < 3) { RAISE_M("Usage: pconvert benchmark <directory> [count]"); }
+    if(argc > 3) { count = atoi(argv[3]); }
 
     for(index = 0; index < ALGORITHMS_SIZE; index++) {
         for(index_j = 0; index_j < COMPRESSION_SIZE; index_j++) {
             for(index_k = 0; index_k < OPENCL_SIZE; index_k++) {
-                time = pbenchmark_algorithm(argv[2], algorithms[index], NULL, "alpha", compression[index_j], 0, use_opencl[index_k]);
-                printf("%s %s %s: %0.2f ms\n", algorithms[index], compression_s[index_j], use_opencl_s[index_k], time * 1000);
+                time = 0.0f;
+                benchmark.blend_time = 0;
+                benchmark.read_png_time = 0;
+                benchmark.write_png_time = 0;
+                sprintf(label, "%s %s %s", algorithms[index], compression_s[index_j], use_opencl_s[index_k]);
+                printf("%-42s ", label);
+                fflush(stdout);
+                for(index_c = 0; index_c < (size_t) count; index_c++) {
+                    result = pbenchmark_algorithm(
+                        argv[2], algorithms[index], NULL, "alpha",
+                        compression[index_j], 0, use_opencl[index_k],
+                        &benchmark
+                    );
+                    time += result;
+                    total += result;
+                    is_success = result >= 0.0f;
+                    if (!is_success) break;
+                }
+                if (is_success) sprintf(time_s, "%0.2fms", time * 1000.0f);
+                else sprintf(time_s, "ERROR!");
+                printf("%s", time_s);
+                if(details == TRUE && is_success == TRUE) {
+                    printf(
+                        " (blend %0.2fms, read %0.2fms, write %0.2fms)",
+                        benchmark.blend_time * 1000.0f,
+                        benchmark.read_png_time * 1000.0f,
+                        benchmark.write_png_time * 1000.0f
+                    );
+                }
+                benchmark_total.blend_time += benchmark.blend_time;
+                benchmark_total.read_png_time += benchmark.read_png_time;
+                benchmark_total.write_png_time += benchmark.read_png_time;
+                printf("\n");
             }
         }
         printf("\n");
     }
 
+    printf("%-42s %0.2fms", "total_execution", total * 1000.0f);
+    if(details == TRUE) {
+        printf(
+            " (blend %0.2fms, read %0.2fms, write %0.2fms)",
+            benchmark_total.blend_time * 1000.0f,
+            benchmark_total.read_png_time * 1000.0f,
+            benchmark_total.write_png_time * 1000.0f
+        );
+    }
+    printf("\n");
+
     NORMAL;
 }
 
@@ -832,26 +955,28 @@ ERROR_T popencl(int argc, char **argv) {
     float end_time;
     float time_elapsed_cpu;
     float time_elapsed_gpu;
+    struct benchmark benchmark_cpu = { 0, 0, 0 };
+    struct benchmark benchmark_gpu = { 0, 0, 0 };
 
     if(argc != 3) { RAISE_M("Usage: pconvert opencl <directory>"); }
 
     time_elapsed_cpu = 0;
     start_time = (float) clock() / CLOCKS_PER_SEC;
-    VALIDATE(compose_images_extra(argv[2], "multiplicative", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, FALSE));
-    VALIDATE(compose_images_extra(argv[2], "source_over", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, FALSE));
-    VALIDATE(compose_images_extra(argv[2], "alpha", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, FALSE));
-    VALIDATE(compose_images_extra(argv[2], "disjoint_over", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, FALSE));
-    VALIDATE(compose_images_extra(argv[2], "disjoint_under", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, FALSE));
+    VALIDATE(compose_images_extra(argv[2], "multiplicative", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, FALSE, &benchmark_cpu));
+    VALIDATE(compose_images_extra(argv[2], "source_over", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, FALSE, &benchmark_cpu));
+    VALIDATE(compose_images_extra(argv[2], "alpha", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, FALSE, &benchmark_cpu));
+    VALIDATE(compose_images_extra(argv[2], "disjoint_over", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, FALSE, &benchmark_cpu));
+    VALIDATE(compose_images_extra(argv[2], "disjoint_under", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, FALSE, &benchmark_cpu));
     end_time = (float) clock() / CLOCKS_PER_SEC;
     time_elapsed_cpu = end_time - start_time;
 
     time_elapsed_gpu = 0;
     start_time = (float) clock() / CLOCKS_PER_SEC;
-    VALIDATE(compose_images_extra(argv[2], "multiplicative", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, TRUE));
-    VALIDATE(compose_images_extra(argv[2], "source_over", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, TRUE));
-    VALIDATE(compose_images_extra(argv[2], "alpha", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, TRUE));
-    VALIDATE(compose_images_extra(argv[2], "disjoint_over", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, TRUE));
-    VALIDATE(compose_images_extra(argv[2], "disjoint_under", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, TRUE));
+    VALIDATE(compose_images_extra(argv[2], "multiplicative", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, TRUE, &benchmark_gpu));
+    VALIDATE(compose_images_extra(argv[2], "source_over", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, TRUE, &benchmark_gpu));
+    VALIDATE(compose_images_extra(argv[2], "alpha", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, TRUE, &benchmark_gpu));
+    VALIDATE(compose_images_extra(argv[2], "disjoint_over", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, TRUE, &benchmark_gpu));
+    VALIDATE(compose_images_extra(argv[2], "disjoint_under", NULL, "alpha", Z_BEST_SPEED, PNG_FILTER_NONE, TRUE, &benchmark_gpu));
     end_time = (float) clock() / CLOCKS_PER_SEC;
     time_elapsed_gpu = end_time - start_time;
 
@@ -864,7 +989,7 @@ ERROR_T pversion(int argc, char **argv) {
     char libpng_version_s[16];
     libpng_version(libpng_version_s);
     printf(
-        "P(NG)Convert %s (%s %s) [%s %s %d bit] [libpng %s] [%s] \n",
+        "P(NG)Convert %s (%s %s) [%s %s %d bit] [libpng %s] [%s]\n",
         PCONVERT_VERSION,
         PCONVERT_COMPILATION_DATE,
         PCONVERT_COMPILATION_TIME,
